/*
 * drivers/staging/gs/fact_cfg.c
 *
 * Driver for the provison.
 *
 * Copyright (C) 2012, 2013 GrandStream Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.
 */

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/delay.h>
#include <linux/rslib.h>
#include <linux/moduleparam.h>
#include <linux/dma-mapping.h>
#include <linux/types.h>
#include <linux/export.h>
#include <linux/semaphore.h>
#include <linux/io.h>
#include <linux/slab.h>
#include <linux/completion.h>
#include <linux/uaccess.h>
#include <linux/proc_fs.h>
#include <linux/vmalloc.h>
#include <linux/module.h>
#include <linux/seq_file.h>
#include <../fs/proc/internal.h>

#include "gs_def.h"
#include <mach/gs_type.h>
#include "flash_rw.h"
#include "mmc_rw.h"

static char *factParaBuf;
static int fact_data_size;
static char *fact_buf;

struct proc_dir_entry *gxvboard_root;

#define EMMC_RWU_SECTORS		(256)

int gvc32xx_getmac_factory(char *mac, unsigned char type);

int factcfg_read(void)
{
	int rval = 0;
	int storage = GS_STORAGE_EMMC;
	sector_t start = FACTORY_SECTOR_OFFSET;
	sector_t total = FACTORY_SECTOR_SIZE;
	sector_t off;
	sector_t nsecs;

	switch (storage) {
	case GS_STORAGE_FLASH:
#ifdef CONFIG_GS_STORAGE_FLASH
	rval = gxv31xx_internal_mtd_read(GS_FACTORY_CFG_OFFSET, GS_FACTORY_CFG_SIZE, factParaBuf);
#endif
		break;
	case GS_STORAGE_EMMC:
#ifdef CONFIG_GS_STORAGE_EMMC
		factParaBuf[0] = 0x55;
		factParaBuf[1] = 0xaa;
		for (off = 0; off < total;) {
			nsecs = total - off;
			if (nsecs > EMMC_RWU_SECTORS)
				nsecs = EMMC_RWU_SECTORS;
			rval = emmc_read(start+off, nsecs, factParaBuf+(off<<9), nsecs<<9);
			if (rval < 0)
				printk("factcfg emmc read err %d", rval);
			off += nsecs;
		}
#endif
		break;
	default:
		break;
	}
	return rval;
}

int factcfg_write(void)
{
	int rval = 0;
	int storage = GS_STORAGE_EMMC;
	sector_t start = FACTORY_SECTOR_OFFSET;
	sector_t total = FACTORY_SECTOR_SIZE;
	sector_t off;
	sector_t nsecs;

	switch (storage) {
	case GS_STORAGE_FLASH:
#ifdef CONFIG_GS_STORAGE_FLASH
		gxv31xx_internal_mtd_erase(GS_FACTORY_CFG_OFFSET, GS_FACTORY_CFG_SIZE);
		rval = gxv31xx_internal_mtd_write(GS_FACTORY_CFG_OFFSET, GS_FACTORY_CFG_SIZE, factParaBuf);
#endif
		break;
	case GS_STORAGE_EMMC:
#ifdef CONFIG_GS_STORAGE_EMMC
		for (off = 0; off < total;) {
			nsecs = total-off;
			if (nsecs > EMMC_RWU_SECTORS)
				nsecs = EMMC_RWU_SECTORS;
			rval = emmc_write(start+off, nsecs, factParaBuf+(off<<9), nsecs<<9);
			if (rval < 0)
				printk("factcfg emmc write err %d", rval);
			off += nsecs;
		}
#endif
		break;
	default:
		break;
	}
	return rval;
}


static unsigned short checksum(GS_FACTORYPARA *factPara)
{
	unsigned char  *p = (unsigned char *)factPara;
	unsigned short sum = 0;
	unsigned short checksum = 0;
	unsigned length = 0;
	int i;
/*the length field is designed to be able expand the structure but since we need copy
  the whole structure, it makes nothing except has chance to crash when length is invalid*/
	length = factPara->length[1]<<16|factPara->length[0];

	if (length > (FACTORY_SECTOR_SIZE<<9))
	length = (FACTORY_SECTOR_SIZE<<9);

	for (i = 0; i < length; i += 2) {
	if (i == offsetof(GS_FACTORYPARA, checksum))
		continue;/*skip checksum itself*/
	sum += ((p[i] << 8) | p[i + 1]);/*fact config use be for checksum calc*/
	}
	checksum = 0xffff - sum + 0x1;
	return ((checksum >> 8) | (checksum << 8));
}

static int read_fact_para(const unsigned int id, GS_FACTORYPARA *buf_ptr, char *data)
{
	int length = 0;
	GS_FACTORYPARA *factPara = buf_ptr;

	if (!buf_ptr)
		return -1;

	length = factPara->length[1] << 16 | factPara->length[0];
	if (length <= sizeof(GS_FACTORYPARA))
		return 0;

	if (factPara->checksum == checksum(factPara)) {
		int count = 0;
		unsigned short  *featurePtr = NULL;
		unsigned short data_len = 0;

		count = factPara->hw_feature_table_word[2];
		featurePtr = (unsigned short *)factPara->hw_feature_table_word+3;

		while (count) {
			if (id == *featurePtr) {
				data_len = (*(featurePtr + 1))<<1;
				memcpy((char *)data, (char *)(featurePtr + 2), data_len);
				return data_len;
			} else {
				featurePtr += *(featurePtr + 1) + 2;
				count--;
			}
		}
	}
	return 0;
}

static int update_fact_para(const unsigned int id, GS_FACTORYPARA *buf_ptr, char *data, unsigned int in_data_len)
{
	int length = 0;
	GS_FACTORYPARA *factPara = buf_ptr;
	if(!buf_ptr) return -1;

	length = factPara->length[1] << 16 | factPara->length[0];
	if (length <= sizeof (GS_FACTORYPARA))
		return 0;

	if (factPara->checksum == checksum(factPara))
	{
		int count = 0;
		unsigned short  *featurePtr = NULL;
		unsigned short data_len=0;

		count = factPara->hw_feature_table_word[2];
		featurePtr = (unsigned short*)factPara->hw_feature_table_word+3;

		while (count)
		{
			if (id == *featurePtr)
			{
				data_len = (*(featurePtr+1))<<1;
				if (in_data_len > data_len)
					in_data_len = data_len;
				memcpy((char*)(featurePtr+2), (char*)data, in_data_len);
				return data_len;
			}
			else
			{
				featurePtr += *(featurePtr+1) + 2;
				count--;
			}
		}
	}
	return 0;
}

int fcfg_get_oled_type(unsigned short *type)
{
	unsigned short val = 0;

	read_fact_para(HW_FEATURE_OLED_TYPE, (GS_FACTORYPARA *)factParaBuf, (char *)&val);
	if (type)
	*type = val;

	return 0;
}
EXPORT_SYMBOL(fcfg_get_oled_type);

static void *gs_seq_start(struct seq_file *s, loff_t *pos)
{
	if (*pos >= fact_data_size)
		return NULL;
	else
		return (void *)(fact_buf + *pos);
}

/*************************************************************************
*          Use seq_file to handle who's data big then 1 page             *
**************************************************************************/
static void *gs_seq_next(struct seq_file *s, void *v, loff_t *pos)
{
	(*pos)++;
	return gs_seq_start(s, pos);
}

static void gs_seq_stop(struct seq_file *s, void *v)
{
	return;
}

static int gs_seq_show(struct seq_file *s, void *v)
{
	char *spos = (char *) v;

	seq_printf(s, "%c", *spos);
	return 0;
}

static struct seq_operations gs_seq_ops = {
	.start = gs_seq_start,
	.next = gs_seq_next,
	.stop = gs_seq_stop,
	.show = gs_seq_show
};


static int data_proc_open(struct inode *inode, struct file *file)
{
	GS_FACTORYPARA *factPara = (GS_FACTORYPARA *)factParaBuf;

	if (checksum(factPara) != factPara->checksum)
		return -EFAULT;

	fact_data_size = factPara->length[1] << 16 | factPara->length[0];
	fact_buf = (char *)factPara;

	return seq_open(file, &gs_seq_ops);
}
#if 0
static int logo_proc_open(struct inode *inode, struct file *file)
{
	pic_buf = vmalloc(MAX_PIC_SIZE);
	if (pic_buf == NULL)
		return -ENOMEM;
	fact_data_size = read_fact_para(HW_FEATURE_LOGODATA, (GS_FACTORYPARA *)factParaBuf, pic_buf);
	fact_buf = pic_buf;

	return seq_open(file, &gs_seq_ops);
}

static int logo_proc_release(struct inode *inode, struct file *file)
{
	vfree(pic_buf);
	pic_buf = NULL;

	return seq_release(inode, file);
}
#endif
static struct file_operations fact_data_proc_ops = {
	.owner = THIS_MODULE,
	.open = data_proc_open,
	.read = seq_read,
	.llseek = seq_lseek,
	.release = seq_release
};
#if 0
static struct file_operations logo_data_proc_ops = {
	.owner = THIS_MODULE,
	.open = logo_proc_open,
	.read = seq_read,
	.llseek = seq_lseek,
	.release = logo_proc_release
};
#endif
/*************************************************************************
* Read functions of factory data for /proc file system
 ************************************************************************/
static int board_dummy_proc_show(struct seq_file *m, void *v)
{
	char buf;
	void *data = (void *)m->private;

	if (((long)data) == 1) {
	read_fact_para(HW_FEATURE_FXO_CS, (GS_FACTORYPARA *)factParaBuf, &buf);
	} else {
	read_fact_para(HW_FEATURE_FXS_CS, (GS_FACTORYPARA *)factParaBuf, &buf);
	}
	seq_printf(m, "%d", buf);

	return 0;
}

#if 0
static int board_factcfg_cam_bits_proc_read(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int len = 0;
	char buf;

	len = read_fact_para(HW_FEATURE_CAM_BITS, (GS_FACTORYPARA *)factParaBuf, &buf);
	sprintf(page, "%d", buf);

	if (len <= off+count)
		*eof = 1;
	*start = page + off;
	len -= off;
	if (len > count)
		len = count;
	if (len < 0)
		len = 0;

	return len;
}


static int board_factcfg_ts_chip_proc_read(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int len = 0;
	char buf;

	len = read_fact_para(HW_FEATURE_TOUCH_SCREEN_TYPE, (GS_FACTORYPARA *)factParaBuf, &buf);
	sprintf(page, "%d", buf);

	if (len <= off+count)
		*eof = 1;
	*start = page + off;
	len -= off;
	if (len > count)
		len = count;
	if (len < 0)
		len = 0;

	return len;
}

static int board_factcfg_usb_gpio_proc_read(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int len = 0;
	char buf;

	len = read_fact_para(HW_FEATURE_USB_GPIO_SET, (GS_FACTORYPARA *)factParaBuf, &buf);
	sprintf(page, "%d", buf);

	if (len <= off+count)
		*eof = 1;
	*start = page + off;
	len -= off;
	if (len > count)
		len = count;
	if (len < 0)
		len = 0;

	return len;
}
#endif

int board_factcfg_hwmodel_proc_show(struct seq_file *m, void *v)
{
	char hwmodel[64];
	void *data = m->private;

	memset(hwmodel, 0, sizeof(hwmodel));
	switch ((long)data) {
	case 0:
	read_fact_para(HW_FEATURE_MODELALIAS, (GS_FACTORYPARA *)factParaBuf, hwmodel);
		break;
	case 1:
	read_fact_para(HW_FEATURE_MODELBASE, (GS_FACTORYPARA *)factParaBuf, hwmodel);
		break;
	case 2:
	read_fact_para(HW_FEATURE_MODELSTR, (GS_FACTORYPARA *)factParaBuf, hwmodel);
		break;
	}
	seq_printf(m, "%s", hwmodel);

	return 0;
}

int board_factcfg_wifi_bt_type_proc_show(struct seq_file *m, void *v)
{
	char type[10];

	memset(type, 0, sizeof(type));
	read_fact_para(HW_FEATURE_WIFI_BT, (GS_FACTORYPARA *)factParaBuf, type);
	seq_printf(m, "%s", type);

	return 0;
}

int board_factcfg_builtin_wifi_bt_proc_show(struct seq_file *m, void *v)
{
	int len = 0;
	char buf;

	len = read_fact_para(HW_FEATURE_BUILT_IN_WIFI, (GS_FACTORYPARA *)factParaBuf, &buf);
	if (len == 0)
		seq_printf(m, "%d", 1);
	else
		seq_printf(m, "%d", buf);

	return 0;
}

int board_factcfg_builtin_usb_proc_show(struct seq_file *m, void *v)
{
	int len = 0;
	char buf;

	len = read_fact_para(HW_FEATURE_BUILT_IN_USB, (GS_FACTORYPARA *)factParaBuf, &buf);
	if (len == 0)
		seq_printf(m, "%d", 1);
	else
		seq_printf(m, "%d", buf);

	return 0;
}

int board_factcfg_hwrev_proc_show(struct seq_file *m, void *v)
{
	int len = 0;
	char hwrev[128];
	GS_FACTORYPARA *factPara = (GS_FACTORYPARA *) factParaBuf;

	memset(hwrev, 0, sizeof(hwrev));
	if (factPara->checksum == checksum(factPara)) {
	char *pn = (char *)factPara->PartNumber;

	sprintf(hwrev, "V%c.%c%c", pn[8], pn[9], pn[10]);
	len = strlen(hwrev);
	}
	seq_printf(m, "%s", hwrev);

	return 0;
}

static int board_factcfg_dev_mac_proc_show(struct seq_file *m, void *v)
{
	GS_FACTORYPARA *factPara = (GS_FACTORYPARA *) factParaBuf;
	int len = 0;
	char *default_mac = "00:0B:82:FF:FF:FF";
	char dev_mac[32];

	memset(dev_mac, 0, sizeof(dev_mac));
	if (factPara->checksum == checksum(factPara)) {
	sprintf(dev_mac, "%02X:%02X:%02X:%02X:%02X:%02X",
		factPara->mac[0] & 0xFF, (factPara->mac[0] >> 8) & 0xFF,
		factPara->mac[1] & 0xFF, (factPara->mac[1] >> 8) & 0xFF,
		factPara->mac[2] & 0xFF, (factPara->mac[2] >> 8) & 0xFF);
		len = strlen(dev_mac);
	} else {
	len = strlen(default_mac);
	memcpy(dev_mac, default_mac, len);
	}
	seq_printf(m, "%s", dev_mac);

	return 0;
}

static ssize_t board_factcfg_dev_mac_proc_write(struct file *file,
		const char __user *buffer, size_t count, loff_t *pos)

{
	GS_FACTORYPARA *factPara = (GS_FACTORYPARA *) factParaBuf;
	unsigned int mac[6];
	unsigned char mac_str[64] = {0};


	if (copy_from_user(mac_str, buffer, count))
		return (-EFAULT);

	sscanf(mac_str, "%02x:%02x:%02x:%02x:%02x:%02x", &mac[0], &mac[1], &mac[2], &mac[3], &mac[4], &mac[5]);
	if (factPara->checksum != checksum(factPara))
		return (-EINVAL);

	factPara->mac[0] = (0xff&mac[0]) | (0xff00&(mac[1] << 8));
	factPara->mac[1] = (0xff&mac[2]) | (0xff00&(mac[3] << 8));
	factPara->mac[2] = (0xff&mac[4]) | (0xff00&(mac[5] << 8));

	factPara->checksum = checksum(factPara);

	factcfg_write();
	return count;
}

static int board_factcfg_wlan_mac_proc_show (struct seq_file *m, void *v)
{
    char dev_mac[32];

	GS_FACTORYPARA* factPara = (GS_FACTORYPARA*) factParaBuf;

	if (!factPara)
		return -EFAULT;
	if (factPara->checksum != checksum(factPara))
		return (-EINVAL);

	memset(dev_mac, 0, sizeof(dev_mac));
	/*if bt mac is not set, try to use ethernet mac + 1*/
	if ((read_fact_para(HW_FEATURE_WLAN_MAC, (GS_FACTORYPARA*)factPara, dev_mac) <= 0)
		|| (strstr(dev_mac, "00:00:00") != NULL)
		|| (strstr(dev_mac, "FF:FF:FF") != NULL)) {
		char mac[6];
		gvc32xx_getmac_factory(mac, 1);
		sprintf(dev_mac, "%02hhx:%02hhx:%02hhx:%02hhx:%02hhx:%02hhx",
			mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]);
	}

	seq_printf(m, "%s", dev_mac);
	return 0;
}

static ssize_t board_factcfg_wlan_mac_proc_write(struct file *file,
		const char __user *buffer, size_t count, loff_t *pos)

{
	GS_FACTORYPARA* factPara = (GS_FACTORYPARA*) factParaBuf;
	unsigned char mac_str[64] = {0};

	memset(mac_str, 0, sizeof(mac_str));
	if (copy_from_user(mac_str, buffer, count))
		return (-EFAULT);
	if (factPara->checksum != checksum(factPara))
		return (-EINVAL);

	update_fact_para(HW_FEATURE_WLAN_MAC, factPara, mac_str, strlen(mac_str));

	if (strcmp(CONFIG_ARCH_MTK_PROJECT, "wp800") == 0) {
		unsigned int wlan_mac[6];
		sscanf(mac_str, "%02x:%02x:%02x:%02x:%02x:%02x", &wlan_mac[0], &wlan_mac[1], &wlan_mac[2], &wlan_mac[3], &wlan_mac[4], &wlan_mac[5]);
		factPara->mac[0] = (0xff&wlan_mac[0]) | (0xff00&(wlan_mac[1] << 8));
		factPara->mac[1] = (0xff&wlan_mac[2]) | (0xff00&(wlan_mac[3] << 8));
		factPara->mac[2] = (0xff&wlan_mac[4]) | (0xff00&(wlan_mac[5] << 8));
	}
	factPara->checksum = checksum(factPara);
	factcfg_write();
	return count;
}

static int board_factcfg_bluetooth_mac_proc_show (struct seq_file *m, void *v)
{
	char dev_mac[32];
	GS_FACTORYPARA* factPara = (GS_FACTORYPARA*) factParaBuf;

	if (!factPara)
		return -EFAULT;
	if (factPara->checksum != checksum(factPara))
		return (-EINVAL);

	memset(dev_mac, 0, sizeof(dev_mac));

	/*if bt mac is not set, try to use ethernet mac + 2*/
	if ((read_fact_para(HW_FEATURE_BT_MAC, (GS_FACTORYPARA*)factPara, dev_mac) <= 0)
		|| (strstr(dev_mac, "00:00:00") != NULL)
		|| (strstr(dev_mac, "FF:FF:FF") != NULL)) {
		char mac[6];
		gvc32xx_getmac_factory(mac, 2);
		sprintf(dev_mac, "%02hhx:%02hhx:%02hhx:%02hhx:%02hhx:%02hhx",
			mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]);
	}

	seq_printf(m, "%s", dev_mac);
	return 0;
}

static ssize_t board_factcfg_bluetooth_mac_proc_write(struct file *file,
		const char __user *buffer, size_t count, loff_t *pos)

{
	GS_FACTORYPARA* factPara = (GS_FACTORYPARA*) factParaBuf;
	unsigned char mac_str[64] = {0};

	memset(mac_str, 0, sizeof(mac_str));
	if (copy_from_user(mac_str, buffer, count))
		return (-EFAULT);
	if (factPara->checksum != checksum(factPara))
		return (-EINVAL);

	update_fact_para(HW_FEATURE_BT_MAC, factPara, mac_str, strlen(mac_str));
	factPara->checksum = checksum(factPara);
	factcfg_write();
	return count;
}

int gvc32xx_get_hw_ver_factory(char *hw_ver)
{
	int len = 0;
	GS_FACTORYPARA *factPara = (GS_FACTORYPARA *) factParaBuf;
	char *hwrev = (char *)hw_ver;
	char revStr[128];

	if (factPara->checksum == checksum(factPara)) {
	char *pn = (char *)factPara->PartNumber;

	sprintf(revStr, "%c%c%c", pn[8], pn[9], pn[10]);
	len = strlen(revStr);
	memcpy(hwrev, revStr, len);
	return 0;
	}
	return -1;
}
EXPORT_SYMBOL(gvc32xx_get_hw_ver_factory);

static char g_hwrev[8];
static int __init hwrev(char *str)
{
	memcpy(g_hwrev, str, strlen(str));
	return 0;
}
early_param("hwrev", hwrev);

/* compare hw version according to action
 * hw_ver: hw version that wanted to be compared
 * action: (0 : <) (1 : ==) (2 : >)
 */
bool gs_hw_ver_check(const char *hw_ver, int action)
{
	int ret;
	char *cur_ver;
	bool retval = false;

	cur_ver = g_hwrev;
	if (strlen(cur_ver) > 0) {
		unsigned int cver, hver;
		ret = kstrtou32(cur_ver, 16, &cver);
		ret |= kstrtou32(hw_ver, 16, &hver);
		if (!ret) {
			retval = 0;
			switch (action) {
			case 0:
				if (cver < hver)
					retval = true;
				break;
			case 1:
				if (cver == hver)
					retval = true;
				break;
			case 2:
				if (cver > hver)
					retval = true;
				break;
			default:
				break;
			}
		}
	}

	return retval;
}
EXPORT_SYMBOL(gs_hw_ver_check);

/**********************************************
@func: ethernet mac   = mac_base
       wlan mac       = mac_base + 1
       bluethooth mac = mac_base + 2
@type: 0 - ethernet  1 - wlan  2 - bt
***********************************************/
int gvc32xx_getmac_factory(char *mac, unsigned char type)
{
	char *default_mac = "00:0B:82:FF:FF:FF";
	int i;
	unsigned int m[6];
	GS_FACTORYPARA *factPara = (GS_FACTORYPARA *) factParaBuf;
	if (factPara && factPara->checksum == checksum(factPara)) {
		unsigned char *mc;
		mc = (unsigned char *)factPara->mac;
		for (i = 0; i < 6; i++) {
			if (*mc != 0xff && *mc != 0) {
				break;
			}
			mc++;
		}
		if (i != 6) {
			unsigned int xxx_mac = 0;
			unsigned char dump_mac[3];
			memcpy(mac, factPara->mac, sizeof(factPara->mac));
			/* adjust the mac address accroding to device type */
			if (type > 2)
				type = 0;

			dump_mac[0] = mac[3];
			dump_mac[1] = mac[4];
			dump_mac[2] = mac[5];

			xxx_mac = dump_mac[2] | (dump_mac[1] << 8) | (dump_mac[0] << 16);
			xxx_mac += type;
			mac[5] = (char)(xxx_mac & 0xFF);
			mac[4] = (char)((xxx_mac >> 8) & 0xFF);
			mac[3] = (char)((xxx_mac >> 16) & 0xFF);
			return 0;
		}
	}
	sscanf(default_mac, "%02x:%02x:%02x:%02x:%02x:%02x", &m[0], &m[1], &m[2], &m[3], &m[4], &m[5]);
	*mac++ = m[0];
	*mac++ = m[1];
	*mac++ = m[2];
	*mac++ = m[3];
	*mac++ = m[4];
	*mac = m[5];
	return 0;
}

EXPORT_SYMBOL(gvc32xx_getmac_factory);

int gvc32xx_get_wlan_mac_factory(char *mac)
{
	char mac_str[32];
	unsigned int m[6];
	GS_FACTORYPARA* factPara = (GS_FACTORYPARA*) factParaBuf;

	if (!factPara)
		return -EFAULT;
	if (factPara->checksum != checksum(factPara))
		return (-EINVAL);

	memset(m, 0, sizeof(m));
	memset(mac_str, 0, sizeof(mac_str));
	/*if bt mac is not set, try to use ethernet mac + 1*/
	if ((read_fact_para(HW_FEATURE_WLAN_MAC, (GS_FACTORYPARA*)factPara, mac_str) <= 0)
		|| (strstr(mac_str, "00:00:00") != NULL)
		|| (strstr(mac_str, "FF:FF:FF") != NULL)) {
		gvc32xx_getmac_factory(mac, 1);
		return 0;
	}

	sscanf(mac_str, "%02x:%02x:%02x:%02x:%02x:%02x", &m[0], &m[1], &m[2], &m[3], &m[4], &m[5]);
	*mac++ = m[0];
	*mac++ = m[1];
	*mac++ = m[2];
	*mac++ = m[3];
	*mac++ = m[4];
	*mac = m[5];
	return 0;
}

EXPORT_SYMBOL(gvc32xx_get_wlan_mac_factory);

int gvc32xx_setmac_factory(const char *mac)
{
	GS_FACTORYPARA *factPara = (GS_FACTORYPARA *) factParaBuf;

	if (!factPara)
		return -1;
	if (factPara->checksum != checksum(factPara)) {
		printk("factory checksum is error\n");
		return -1;
	}
	if (memcmp((char *)factPara->mac, mac, 6) != 0) {
		printk("write mac to flash\n");
		factPara->mac[0] = mac[0] | (mac[1] << 8);
		factPara->mac[1] = mac[2] | (mac[3] << 8);
		factPara->mac[2] = mac[4] | (mac[5] << 8);

		factPara->checksum = checksum(factPara);
	}
	return 0;
}

EXPORT_SYMBOL(gvc32xx_setmac_factory);

static int board_factcfg_ccode_proc_show (struct seq_file *m, void *v)
{
	char ccode[32];
	GS_FACTORYPARA* factPara = (GS_FACTORYPARA*) factParaBuf;

	if (!factPara)
		return -EFAULT;
	if (factPara->checksum != checksum(factPara))
		return (-EINVAL);

	memset(ccode, 0, sizeof(ccode));

	if (read_fact_para(HW_FEATURE_COUNTRY_CODE, (GS_FACTORYPARA*)factPara, ccode) > 0)
		seq_printf(m, "%s", ccode);

	return 0;
}

static ssize_t board_factcfg_load_factcfg_proc_write(struct file *file,
		const char __user *buffer, size_t count, loff_t *pos)
{
	GS_FACTORYPARA *factPara = (GS_FACTORYPARA *) factParaBuf;
	GS_FACTORYPARA *fact_new = vzalloc(FACTORY_SECTOR_SIZE<<9);

	if (!fact_new) {
		printk(KERN_ERR "No mem\n");
		return -ENOMEM;
	}

	if (copy_from_user(fact_new, buffer, count))
		return (-EFAULT);

	if (fact_new->checksum != checksum(fact_new)) {
		printk(KERN_ERR "Invalid factcfg file\n");
		return (-EINVAL);
	}

	if (factPara->checksum == checksum(factPara)) {
		printk(KERN_WARNING "factcfg has been flashed\n");
		return count;
	}

	memcpy(factPara, fact_new, count);
	factcfg_write();
	vfree(fact_new);
	return count;
}

static int board_factcfg_pn_proc_show(struct seq_file *m, void *v)
{
	GS_FACTORYPARA *factPara = (GS_FACTORYPARA *) factParaBuf;
	int len = 0;
	char dev_pn[64];

	memset(dev_pn, 0, sizeof(dev_pn));
	if (factPara->checksum == checksum(factPara)) {
	char *pn = (char *)factPara->PartNumber;
	len = strlen(pn);
	memcpy(dev_pn, pn, len);
	}
	seq_printf(m, "%s", dev_pn);

	return 0;
}

static int board_factcfg_required_bits_proc_show(struct seq_file *m, void *v)
{
	GS_FACTORYPARA *factPara = (GS_FACTORYPARA *) factParaBuf;
	unsigned short required_bits[4];

	memset(required_bits, 0, sizeof(required_bits));
	if (factPara->checksum == checksum(factPara)) {
		required_bits[0] = factPara->RequiredBits[0];
		required_bits[1] = factPara->RequiredBits[1];
		required_bits[2] = factPara->RequiredBits[2];
		required_bits[3] = factPara->RequiredBits[3];

		required_bits[0] = ((required_bits[0] & 0x00FF) << 8) | ((required_bits[0] & 0xFF00) >> 8);
		required_bits[1] = ((required_bits[1] & 0x00FF) << 8) | ((required_bits[1] & 0xFF00) >> 8);
		required_bits[2] = ((required_bits[2] & 0x00FF) << 8) | ((required_bits[2] & 0xFF00) >> 8);
		required_bits[3] = ((required_bits[3] & 0x00FF) << 8) | ((required_bits[3] & 0xFF00) >> 8);
	}
	seq_printf(m, "%04x%04x%04x%04x", required_bits[0], required_bits[1], required_bits[2], required_bits[3]);

	return 0;
}

static int board_factcfg_oemid_proc_show(struct seq_file *m, void *v)
{
	char oemid[16];
	int len;

	memset(oemid, 0 , sizeof(oemid));
	len = read_fact_para(HW_FEATURE_OEM_ID, (GS_FACTORYPARA *)factParaBuf, oemid);
	if (len > 0)
		seq_printf(m, "%d", oemid[0]);

	return 0;
}

static int board_factcfg_fwkey_proc_show(struct seq_file *m, void *v)
{
	char fwkey[128];

	memset(fwkey, 0, sizeof(fwkey));
	read_fact_para(HW_FEATURE_FWKEY, (GS_FACTORYPARA *)factParaBuf, fwkey);
	seq_printf(m, "%s", fwkey);

	return 0;
}

static int board_factcfg_vendor_proc_show(struct seq_file *m, void *v)
{
	char vendor[128];
	void *data = (void *)m->private;

	memset(vendor, 0 , sizeof(vendor));
	if (((long)data) == 0)
		read_fact_para(HW_FEATURE_VENDOR, (GS_FACTORYPARA *)factParaBuf, vendor);
	else
		read_fact_para(HW_FEATURE_VENDOR_FULLNAME, (GS_FACTORYPARA *)factParaBuf, vendor);


	seq_printf(m, "%s", vendor);

	return 0;
}
#if 0
static int board_factcfg_firmware_proc_read(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int len = 0;
	char buf;

	len = read_fact_para(HW_FEATURE_LOWEST_FIRMWARE_VERSION, (GS_FACTORYPARA *)factParaBuf, &buf);
	sprintf(page, "%d", buf);

	if (len <= off+count)
		*eof = 1;
	*start = page + off;
	len -= off;
	if (len > count)
		len = count;
	if (len < 0)
		len = 0;

	return len;
}

static int board_factcfg_headset_proc_read(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int len = 0;
	char buf;

	len = read_fact_para(HW_FEATURE_HEADSET_TYPE, (GS_FACTORYPARA *)factParaBuf, &buf);
	sprintf(page, "%d", buf);

	if (len <= off+count)
		*eof = 1;
	*start = page + off;
	len -= off;
	if (len > count)
		len = count;
	if (len < 0)
		len = 0;

	return len;
}

static int board_factcfg_eth_proc_read(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int len = 0;
	char buf;

	len = read_fact_para(HW_FEATURE_ETH_DEV, (GS_FACTORYPARA *)factParaBuf, &buf);
	sprintf(page, "%d", buf);

	if (len <= off+count)
		*eof = 1;
	*start = page + off;
	len -= off;
	if (len > count)
		len = count;
	if (len < 0)
		len = 0;

	return len;
}

static int board_factcfg_flash_proc_read(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int len = 0;
	char buf;

	len = read_fact_para(HW_FEATURE_FLASH_TYPE, (GS_FACTORYPARA *)factParaBuf, &buf);
	sprintf(page, "%d", buf);

	if (len <= off+count)
		*eof = 1;
	*start = page + off;
	len -= off;
	if (len > count)
		len = count;
	if (len < 0)
		len = 0;

	return len;
}

static int board_factcfg_cam_sensor_proc_read(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int len = 0;
	char buf;

	len = read_fact_para(HW_FEATURE_CAMERA_SENSOR_TYPE, (GS_FACTORYPARA *)factParaBuf, &buf);
	sprintf(page, "%d", buf);

	if (len <= off+count)
		*eof = 1;
	*start = page + off;
	len -= off;
	if (len > count)
		len = count;
	if (len < 0)
		len = 0;

	return len;
}
#endif
static int board_factcfg_secs_proc_show(struct seq_file *m, void *v)
{
	void *data = (void *)m->private;
	int index = (int)((long)data);
	int len = 0;
	char buf;

	len = read_fact_para(HW_FEATURE_REQUIRE_PEM + index, (GS_FACTORYPARA *)factParaBuf, &buf);
	if (len == 0) {
		len = 1;
		if (index == 0)
			seq_printf(m, "%d", 1);
		else
			seq_printf(m, "%d", 0);
	} else
		seq_printf(m, "%d", buf);

	return 0;
}

static int board_factcfg_load_factcfg_proc_show(struct seq_file *m, void *v)
{
	return 0;
}
#if 0
static int board_factcfg_cdc_proc_read(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int len = 0;
	char buf;

	len = read_fact_para(HW_FEATURE_CDC_DEV, (GS_FACTORYPARA *)factParaBuf, &buf);
	sprintf(page, "%d", buf);

	if (len <= off+count)
		*eof = 1;
	*start = page + off;
	len -= off;
	if (len > count)
		len = count;
	if (len < 0)
		len = 0;

	return len;
}

static int board_factcfg_shutdown_proc_read(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int len = 0;
	char buf;

	len = read_fact_para(HW_FEATURE_PA_SHUTDOWN_TYPE, (GS_FACTORYPARA *)factParaBuf, &buf);
	sprintf(page, "%d", buf);

	if (len <= off+count)
		*eof = 1;
	*start = page + off;
	len -= off;
	if (len > count)
		len = count;
	if (len < 0)
		len = 0;

	return len;
}

static int board_factcfg_cam_standby_proc_read(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int len = 0;
	char buf;

	len = read_fact_para(HW_FEATURE_CAMERA_STANDBY_CTRL, (GS_FACTORYPARA *)factParaBuf, &buf);
	sprintf(page, "%d", buf);

	if (len <= off+count)
		*eof = 1;
	*start = page + off;
	len -= off;
	if (len > count)
		len = count;
	if (len < 0)
		len = 0;

	return len;
}

static int board_factcfg_cpu_proc_read(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int len = 0;
	char buf;

	len = read_fact_para(HW_FEATURE_CPU_TYPE, (GS_FACTORYPARA *)factParaBuf, &buf);
	sprintf(page, "%d", buf);

	if (len <= off+count)
		*eof = 1;
	*start = page + off;
	len -= off;
	if (len > count)
		len = count;
	if (len < 0)
		len = 0;

	return len;
}

static int board_factcfg_gain_proc_read(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int len = 0;
	char buf;

	len = read_fact_para(HW_FEATURE_HF_PA_GAIN, (GS_FACTORYPARA *)factParaBuf, &buf);
	sprintf(page, "%d", buf);

	if (len <= off+count)
		*eof = 1;
	*start = page + off;
	len -= off;
	if (len > count)
		len = count;
	if (len < 0)
		len = 0;

	return len;
}

static int board_factcfg_speaker_proc_read(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int len = 0;
	char buf;

	len = read_fact_para(HW_FEATURE_HF_SPEAKER_TYPE, (GS_FACTORYPARA *)factParaBuf, &buf);
	sprintf(page, "%d", buf);

	if (len <= off+count)
		*eof = 1;
	*start = page + off;
	len -= off;
	if (len > count)
		len = count;
	if (len < 0)
		len = 0;

	return len;
}

static int board_factcfg_mic_proc_read(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int len = 0;
	char buf;

	len = read_fact_para(HW_FEATURE_HF_MIC_TYPE, (GS_FACTORYPARA *)factParaBuf, &buf);
	sprintf(page, "%d", buf);

	if (len <= off+count)
		*eof = 1;
	*start = page + off;
	len -= off;
	if (len > count)
		len = count;
	if (len < 0)
		len = 0;

	return len;
}

static int board_factcfg_handset_proc_read(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int len = 0;
	char buf;

	len = read_fact_para(HW_FEATURE_HANDSET_TYPE, (GS_FACTORYPARA *)factParaBuf, &buf);
	sprintf(page, "%d", buf);

	if (len <= off+count)
		*eof = 1;
	*start = page + off;
	len -= off;
	if (len > count)
		len = count;
	if (len < 0)
		len = 0;

	return len;
}

static int board_factcfg_audio_proc_read(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int len = 0;
	char buf;

	len = read_fact_para(HW_FEATURE_AUDIO_CDC_MCLK_RATE, (GS_FACTORYPARA *)factParaBuf, &buf);
	sprintf(page, "%d", buf);

	if (len <= off+count)
		*eof = 1;
	*start = page + off;
	len -= off;
	if (len > count)
		len = count;
	if (len < 0)
		len = 0;

	return len;
}

static int board_factcfg_usb_hub_proc_read(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int len = 0;
	char buf;

	len = read_fact_para(HW_FEATURE_USB_HUB_CLK_SRC, (GS_FACTORYPARA *)factParaBuf, &buf);
	sprintf(page, "%d", buf);

	if (len <= off+count)
		*eof = 1;
	*start = page + off;
	len -= off;
	if (len > count)
		len = count;
	if (len < 0)
		len = 0;

	return len;
}
#endif

static int sec_feature_proc_open(struct inode *inode, struct file *file)
{
	unsigned short *featurePtr, idx;
	unsigned int *feature_table_len;
	GS_FACTORYPARA *factPara = (GS_FACTORYPARA *)factParaBuf;

	void *data = PDE(inode)->data;
	unsigned int delta, fact_len;
    /* It's not neccessary to refresh, so comment */
#if 0
    /* refresh factParaBuf */
	if (factcfg_read()) {
	pr_err("factcfg_read failed!\n");
	return 0;
	}
#endif
	fact_data_size = 0;
	featurePtr = factPara->hw_feature_table_word;
	fact_len = factPara->length[1]<<16|factPara->length[0];

	do {
	feature_table_len = (unsigned int *)featurePtr;
	featurePtr += *feature_table_len;
	delta = (unsigned int)((void *)featurePtr - (void *)factParaBuf);
	if ((delta >= fact_len) || (delta >= GS_FACTORY_CFG_SIZE))
		goto no_sec_table;
	} while (HW_SECURITY_TABLE_MAGIC != *(featurePtr + 1));

	idx = *(featurePtr + 2);
	featurePtr += 3;

	while (idx) {
	if (((void *)featurePtr - (void *)factParaBuf) > GS_FACTORY_CFG_SIZE)
		return 0;

	if ((long)data == *featurePtr) {
		fact_data_size = *(featurePtr+1);
		fact_buf = (char *)(featurePtr + 2);
		break;
	} else {
		featurePtr += ((*(featurePtr + 1) + 1) >> 1) + 2;
		idx--;
	}
	}
no_sec_table:
	pr_info("No security table!\n");

	return seq_open(file, &gs_seq_ops);;
}

static ssize_t sec_feature_proc_write(struct file *file, const char __user *buffer,
	size_t count, loff_t *pos)
{
	GS_FACTORYPARA *factPara = (GS_FACTORYPARA *) factParaBuf;
	unsigned short *featurePtr, *sec_ptr, idx;
	unsigned short sec_table_size = 0, sec_table_entries = 0;
	unsigned char *existing_features;
	unsigned int *feature_table_len, factpara_len = 0;
	void *data = PDE(file->f_path.dentry->d_inode)->data;

	if (factPara->checksum != checksum(factPara))
		return -EINVAL;

	featurePtr = factPara->hw_feature_table_word;
	factpara_len = sizeof(GS_FACTORYPARA);
	while (HW_SECURITY_TABLE_MAGIC != *(featurePtr + 1)) {
	feature_table_len = (unsigned int *)featurePtr;
	featurePtr += *feature_table_len;
	factpara_len += *feature_table_len << 1;
	if ((HW_SECURITY_TABLE_MAGIC != *(featurePtr + 1)) &&
		(0x4c47 != *(featurePtr + 3))) {
		*(featurePtr + 1) = HW_SECURITY_TABLE_MAGIC;
		*(featurePtr + 2) = 0;
		break;
	}
	}

	idx = *(featurePtr + 2);

	existing_features = vmalloc(PAGE_SIZE * 4);
	if (existing_features == NULL)
	return -ENOMEM;

    /* save existing entries */
	if (idx > 0) {
	sec_ptr = featurePtr + 3;

	while (idx) {

		if ((long)data != *sec_ptr) {
		int ft_len = *(sec_ptr + 1) + 2 * sizeof(unsigned short);
		if (ft_len + sec_table_size > PAGE_SIZE * 4)
			break;
		memcpy(existing_features + sec_table_size, (void *)sec_ptr, ft_len);
		sec_table_size += ((ft_len + 1) >> 1) * 2;
		sec_table_entries++;
		sec_ptr += ((*(sec_ptr + 1) + 1) >> 1) + 2;
		idx--;
		} else /* we already have one in flash, exit */
		goto finish;
		}
	}
	/* copy new entry */
	if (sec_table_size + count > PAGE_SIZE * 4)
		return -EINVAL;
	else {
		sec_ptr = (unsigned short *)(existing_features + sec_table_size);
		*sec_ptr = (long)data;
		*(sec_ptr + 1) = count;
		if (copy_from_user((void *)(sec_ptr + 2), buffer, count))
			return -EFAULT;
		/* 2 bytes align */
		sec_table_size += ((count + 1) >> 1) * 2 + 2 * sizeof(unsigned short);
		sec_table_entries++;
	}

	/* fill table */
	memcpy((void *)(featurePtr + 3), existing_features, sec_table_size);
	sec_table_size += 3 * sizeof(unsigned short);
	*featurePtr = sec_table_size >> 1;
	*(featurePtr + 2) = sec_table_entries;

	factpara_len += sec_table_size;
	factPara->length[0] = (unsigned short)factpara_len;
	factPara->length[1] = (unsigned short)(factpara_len >> 16);
	factPara->checksum = checksum(factPara);

	factcfg_write();
finish:
	vfree(existing_features);
	return count;
}

static struct file_operations sec_proc_ops = {
	.owner = THIS_MODULE,
	.open = sec_feature_proc_open,
	.read = seq_read,
	.write = sec_feature_proc_write,
	.llseek = seq_lseek,
	.release = seq_release,
};


static int board_factcfg_hwmodel_proc_open(struct inode *inode, struct file *file)
{
	return single_open(file, board_factcfg_hwmodel_proc_show, PDE_DATA(inode));
}

static int board_factcfg_wifi_bt_type_proc_open(struct inode *inode, struct file *file)
{
	return single_open(file, board_factcfg_wifi_bt_type_proc_show, PDE_DATA(inode));
}

static int board_factcfg_builtin_wifi_bt_proc_open(struct inode *inode, struct file *file)
{
	return single_open(file, board_factcfg_builtin_wifi_bt_proc_show, PDE_DATA(inode));
}

static int board_factcfg_builtin_usb_proc_open(struct inode *inode, struct file *file)
{
	return single_open(file, board_factcfg_builtin_usb_proc_show, PDE_DATA(inode));
}

static const struct file_operations board_factcfg_hwmodel_proc_fops = {
	.open  = board_factcfg_hwmodel_proc_open,
	.read  = seq_read,
	.release = single_release,
	.owner = THIS_MODULE,
};

static const struct file_operations board_factcfg_wifi_bt_type_proc_fops = {
	.open  = board_factcfg_wifi_bt_type_proc_open,
	.read  = seq_read,
	.release = single_release,
	.owner = THIS_MODULE,
};

static const struct file_operations board_factcfg_builtin_wifi_bt_proc_fops = {
	.open  = board_factcfg_builtin_wifi_bt_proc_open,
	.read  = seq_read,
	.release = single_release,
	.owner = THIS_MODULE,
};

static const struct file_operations board_factcfg_builtin_usb_proc_fops = {
	.open  = board_factcfg_builtin_usb_proc_open,
	.read  = seq_read,
	.release = single_release,
	.owner = THIS_MODULE,
};

static int board_factcfg_hwrev_proc_open(struct inode *inode, struct file *file)
{
	return single_open(file, board_factcfg_hwrev_proc_show, PDE_DATA(inode));
}

static const struct file_operations board_factcfg_hwrev_proc_fops = {
	.open  = board_factcfg_hwrev_proc_open,
	.read  = seq_read,
	.release = single_release,
	.owner = THIS_MODULE,
};

static int board_factcfg_fwkey_proc_open(struct inode *inode, struct file *file)
{
	return single_open(file, board_factcfg_fwkey_proc_show, PDE_DATA(inode));
}

static const struct file_operations board_factcfg_fwkey_proc_fops = {
	.open  = board_factcfg_fwkey_proc_open,
	.read  = seq_read,
	.release = single_release,
	.owner = THIS_MODULE,
};

static int board_factcfg_vendor_proc_open(struct inode *inode, struct file *file)
{
	return single_open(file, board_factcfg_vendor_proc_show, PDE_DATA(inode));
}

static const struct file_operations board_factcfg_vendor_proc_fops = {
	.open  = board_factcfg_vendor_proc_open,
	.read  = seq_read,
	.release = single_release,
	.owner = THIS_MODULE,
};


static int board_factcfg_oemid_proc_open(struct inode *inode, struct file *file)
{
	return single_open(file, board_factcfg_oemid_proc_show, PDE_DATA(inode));
}

static const struct file_operations board_factcfg_oemid_proc_fops = {
	.open  = board_factcfg_oemid_proc_open,
	.read  = seq_read,
	.release = single_release,
	.owner = THIS_MODULE,
};

static int board_factcfg_dev_mac_proc_open(struct inode *inode, struct file *file)
{
	return single_open(file, board_factcfg_dev_mac_proc_show, PDE_DATA(inode));
}

static const struct file_operations board_factcfg_dev_mac_proc_fops = {
	.open  = board_factcfg_dev_mac_proc_open,
	.read  = seq_read,
	.write = board_factcfg_dev_mac_proc_write,
	.release = single_release,
	.owner = THIS_MODULE,
};

static int board_factcfg_wlan_mac_proc_open(struct inode *inode, struct file *file)
{
	return single_open(file, board_factcfg_wlan_mac_proc_show, PDE_DATA(inode));
}

static const struct file_operations board_factcfg_wlan_mac_proc_fops = {
	.open  = board_factcfg_wlan_mac_proc_open,
	.read  = seq_read,
	.write = board_factcfg_wlan_mac_proc_write,
	.release = single_release,
	.owner = THIS_MODULE,
};

static int board_factcfg_bluetooth_mac_proc_open(struct inode *inode, struct file *file)
{
	return single_open(file, board_factcfg_bluetooth_mac_proc_show, PDE_DATA(inode));
}

static const struct file_operations board_factcfg_bluetooth_mac_proc_fops = {
	.open  = board_factcfg_bluetooth_mac_proc_open,
	.read  = seq_read,
	.write = board_factcfg_bluetooth_mac_proc_write,
	.release = single_release,
	.owner = THIS_MODULE,
};

static int board_factcfg_ccode_proc_open(struct inode *inode, struct file *file)
{
	return single_open(file, board_factcfg_ccode_proc_show, PDE_DATA(inode));
}

static const struct file_operations board_factcfg_ccode_proc_fops = {
	.open  = board_factcfg_ccode_proc_open,
	.read  = seq_read,
	.release = single_release,
	.owner = THIS_MODULE,
};

static int board_factcfg_pn_proc_open(struct inode *inode, struct file *file)
{
	return single_open(file, board_factcfg_pn_proc_show, PDE_DATA(inode));
}

static const struct file_operations board_factcfg_pn_proc_fops = {
	.open  = board_factcfg_pn_proc_open,
	.read  = seq_read,
	.release = single_release,
	.owner = THIS_MODULE,
};

static int board_factcfg_required_bits_proc_open(struct inode *inode, struct file *file)
{
	return single_open(file, board_factcfg_required_bits_proc_show, PDE_DATA(inode));
}

static const struct file_operations board_factcfg_required_bits_proc_fops = {
	.open  = board_factcfg_required_bits_proc_open,
	.read  = seq_read,
	.release = single_release,
	.owner = THIS_MODULE,
};

static int board_dummy_proc_open(struct inode *inode, struct file *file)
{
	return single_open(file, board_dummy_proc_show, PDE_DATA(inode));
}

static const struct file_operations board_dummy_proc_fops = {
	.open  = board_dummy_proc_open,
	.read  = seq_read,
	.release = single_release,
	.owner = THIS_MODULE,
};

static int board_factcfg_secs_proc_open(struct inode *inode, struct file *file)
{
	return single_open(file, board_factcfg_secs_proc_show, PDE_DATA(inode));
}

static const struct file_operations board_factcfg_secs_proc_fops = {
	.open  = board_factcfg_secs_proc_open,
	.read  = seq_read,
	.release = single_release,
	.owner = THIS_MODULE,
};

static int board_factcfg_load_factcfg_proc_open(struct inode *inode, struct file *file)
{
	return single_open(file, board_factcfg_load_factcfg_proc_show, PDE_DATA(inode));
}

static const struct file_operations board_factcfg_load_factcfg_proc_fops = {
	.open  = board_factcfg_load_factcfg_proc_open,
	.read  = seq_read,
	.write = board_factcfg_load_factcfg_proc_write,
	.release = single_release,
	.owner = THIS_MODULE,
};


/*************************************************************************
 *Install the /proc entry for factory data
 ************************************************************************/
static struct proc_dir_entry *install_proc_entry(void)
{
	struct proc_dir_entry *gxvboard_root_dir;
	struct proc_dir_entry *ent, *sent;
	struct proc_dir_entry *dent = NULL;

	gxvboard_root_dir = proc_mkdir("gsboard", NULL);
	if (gxvboard_root_dir) {
		dent = proc_mkdir("dev_info", gxvboard_root_dir);
	}

	if (dent != NULL) {
	ent = proc_create_data("dev_data", S_IFREG|S_IRUGO, dent, &fact_data_proc_ops, NULL);

	ent = proc_create_data("dev_alias", S_IFREG|S_IRUGO, dent, &board_factcfg_hwmodel_proc_fops, (void *)0);

	ent = proc_create_data("dev_base", S_IFREG|S_IRUGO, dent, &board_factcfg_hwmodel_proc_fops, (void *)1);

	ent = proc_create_data("dev_id", S_IFREG|S_IRUGO, dent, &board_factcfg_hwmodel_proc_fops, (void *)2);

	ent = proc_create_data("dev_rev", S_IFREG|S_IRUGO, dent, &board_factcfg_hwrev_proc_fops, NULL);

	ent = proc_create_data("fwkey", S_IFREG|S_IRUGO, dent, &board_factcfg_fwkey_proc_fops, NULL);

	ent = proc_create_data("wifi_bt_type", S_IFREG|S_IRUGO, dent, &board_factcfg_wifi_bt_type_proc_fops, NULL);
	ent = proc_create_data("builtin_wifi", S_IFREG|S_IRUGO, dent, &board_factcfg_builtin_wifi_bt_proc_fops, NULL);
	ent = proc_create_data("builtin_usb", S_IFREG|S_IRUGO, dent, &board_factcfg_builtin_usb_proc_fops, NULL);

		/*ent = proc_create_data("logo", S_IFREG|S_IRUGO, dent, &logo_data_proc_ops, NULL);*/

	ent = proc_create_data("vendor", S_IFREG|S_IRUGO, dent, &board_factcfg_vendor_proc_fops, (void *)0);

	ent = proc_create_data("vendor_fullname", S_IFREG|S_IRUGO, dent, &board_factcfg_vendor_proc_fops, (void *)1);

	ent = proc_create_data("oem_id", S_IFREG|S_IRUGO, dent, &board_factcfg_oemid_proc_fops, NULL);

	ent = proc_create_data("dev_mac", S_IFREG|S_IRUGO, dent, &board_factcfg_dev_mac_proc_fops, NULL);

	ent = proc_create_data("wlan_mac", S_IFREG|S_IRUGO, dent, &board_factcfg_wlan_mac_proc_fops, NULL);

	ent = proc_create_data("bluetooth_mac", S_IFREG|S_IRUGO, dent, &board_factcfg_bluetooth_mac_proc_fops, NULL);

	ent = proc_create_data("country_code", S_IFREG|S_IRUGO, dent, &board_factcfg_ccode_proc_fops, NULL);

	ent = proc_create_data("PN", S_IFREG|S_IRUGO, dent, &board_factcfg_pn_proc_fops, NULL);

	ent = proc_create_data("required_bits", S_IFREG|S_IRUGO, dent, &board_factcfg_required_bits_proc_fops, NULL);

	ent = proc_create_data("have_FXS", S_IFREG|S_IRUGO, dent, &board_dummy_proc_fops, (void *)0);

	ent = proc_create_data("have_FXO", S_IFREG|S_IRUGO, dent, &board_dummy_proc_fops, (void *)1);

		ent = proc_create_data("require_pem", S_IFREG|S_IRUGO, dent, &board_factcfg_secs_proc_fops, (void *)0);

		ent = proc_create_data("require_sn", S_IFREG|S_IRUGO, dent, &board_factcfg_secs_proc_fops, (void *)1);

		ent = proc_create_data("require_admin_pwd", S_IFREG|S_IRUGO, dent, &board_factcfg_secs_proc_fops, (void *)2);

		ent = proc_create_data("require_xml_key", S_IFREG|S_IRUGO, dent, &board_factcfg_secs_proc_fops, (void *)3);

		ent = proc_create_data("require_hash", S_IFREG|S_IRUGO, dent, &board_factcfg_secs_proc_fops, (void *)4);
#if 0
		ent = create_proc_entry("cam_bits", S_IFREG|S_IRUGO, dent);
		if (ent) {
			ent->read_proc = board_factcfg_cam_bits_proc_read;
			ent->data = NULL;
		}

		ent = create_proc_entry("ts_chip_type", S_IFREG|S_IRUGO, dent);
		if (ent) {
			ent->read_proc = board_factcfg_ts_chip_proc_read;
			ent->data = NULL;
		}

		ent = create_proc_entry("usb_gpio_set", S_IFREG|S_IRUGO, dent);
		if (ent) {
			ent->read_proc = board_factcfg_usb_gpio_proc_read;
			ent->data = NULL;
		}

		ent = create_proc_entry("usb_hub_clk", S_IFREG|S_IRUGO, dent);
		if (ent) {
			ent->read_proc = board_factcfg_usb_hub_proc_read;
			ent->data = NULL;
		}

		ent = create_proc_entry("firmware", S_IFREG|S_IRUGO, dent);
	if (ent) {
		ent->read_proc  = board_factcfg_firmware_proc_read;
		ent->data       = NULL;
	}

	ent = create_proc_entry("headset", S_IFREG|S_IRUGO, dent);
	if (ent) {
		ent->read_proc  = board_factcfg_headset_proc_read;
		ent->data       = NULL;
	}

	ent = create_proc_entry("eth_dev", S_IFREG|S_IRUGO, dent);
	if (ent) {
		ent->read_proc  = board_factcfg_eth_proc_read;
		ent->data       = NULL;
	}

	ent = create_proc_entry("flash", S_IFREG|S_IRUGO, dent);
	if (ent) {
		ent->read_proc  = board_factcfg_flash_proc_read;
		ent->data       = NULL;
	}

	ent = create_proc_entry("cam_sensor", S_IFREG|S_IRUGO, dent);
	if (ent) {
		ent->read_proc  = board_factcfg_cam_sensor_proc_read;
		ent->data       = NULL;
	}

	ent = create_proc_entry("cdc", S_IFREG|S_IRUGO, dent);
	if (ent) {
		ent->read_proc  = board_factcfg_cdc_proc_read;
		ent->data       = NULL;
	}

	ent = create_proc_entry("shutdown", S_IFREG|S_IRUGO, dent);
	if (ent) {
		ent->read_proc  = board_factcfg_shutdown_proc_read;
		ent->data       = NULL;
	}

	ent = create_proc_entry("cam_standby", S_IFREG|S_IRUGO, dent);
	if (ent) {
		ent->read_proc  = board_factcfg_cam_standby_proc_read;
		ent->data       = NULL;
	}

	ent = create_proc_entry("cpu", S_IFREG|S_IRUGO, dent);
	if (ent) {
		ent->read_proc  = board_factcfg_cpu_proc_read;
		ent->data       = NULL;
	}

	ent = create_proc_entry("gain", S_IFREG|S_IRUGO, dent);
	if (ent) {
		ent->read_proc  = board_factcfg_gain_proc_read;
		ent->data       = NULL;
	}

	ent = create_proc_entry("speaker", S_IFREG|S_IRUGO, dent);
	if (ent) {
		ent->read_proc  = board_factcfg_speaker_proc_read;
		ent->data       = NULL;
	}

	ent = create_proc_entry("mic", S_IFREG|S_IRUGO, dent);
	if (ent) {
		ent->read_proc  = board_factcfg_mic_proc_read;
		ent->data       = NULL;
	}

	ent = create_proc_entry("handset", S_IFREG|S_IRUGO, dent);
	if (ent) {
		ent->read_proc  = board_factcfg_handset_proc_read;
		ent->data       = NULL;
	}

	ent = create_proc_entry("audio_rate", S_IFREG|S_IRUGO, dent);
	if (ent) {
		ent->read_proc  = board_factcfg_audio_proc_read;
		ent->data       = NULL;
	}
#endif
		ent = proc_create_data("flash_factcfg", S_IFREG|S_IRUGO, dent, &board_factcfg_load_factcfg_proc_fops, NULL);

		sent = proc_mkdir("security", dent);
		if (sent) {
			ent = proc_create_data("private_key", S_IFREG|S_IRUGO|S_IWUGO, sent, &sec_proc_ops, (void *)0);

			ent = proc_create_data("certificate", S_IFREG|S_IRUGO|S_IWUGO, sent, &sec_proc_ops, (void *)1);

			ent = proc_create_data("checksum", S_IFREG|S_IRUGO|S_IWUGO, sent, &sec_proc_ops, (void *)2);

			ent = proc_create_data("country_code", S_IFREG|S_IRUGO|S_IWUGO, sent, &sec_proc_ops, (void *)3);

			ent = proc_create_data("sn", S_IFREG|S_IRUGO|S_IWUGO, sent, &sec_proc_ops, (void *)4);

			ent = proc_create_data("password", S_IFREG|S_IRUGO|S_IWUGO, sent, &sec_proc_ops, (void *)5);

			ent = proc_create_data("xml_key", S_IFREG|S_IRUGO|S_IWUGO, sent, &sec_proc_ops, (void *)6);

			ent = proc_create_data("hash", S_IFREG|S_IRUGO|S_IWUGO, sent, &sec_proc_ops, (void *)7);
		}
	} else {
		printk("Create proc dir error!\n");
	}

	return dent;
}

static int __init
factcfg_init(void)
{
	int rval;
	GS_FACTORYPARA *factPara;

	factParaBuf = vmalloc(FACTORY_SECTOR_SIZE<<9);
	if (!factParaBuf) {
		printk("factcfg read out of memory\n");
		return -ENOMEM;
	}

	rval = factcfg_read();
	if (rval)
		printk("Load factory data failed!\n");
	else {
		factPara = (GS_FACTORYPARA *) factParaBuf;
		if ((factPara->checksum == checksum(factPara)) && (factPara->checksum > 0))
			printk("factory data loaded.\n");
		else
			printk("factory data invalid checksum.\n");
	}
	if (-ENOMEM != rval)
		gxvboard_root = install_proc_entry();

	return 0;
}

static void __exit
factcfg_cleanup(void)
{
	if (factParaBuf) {
		vfree(factParaBuf);
		factParaBuf = NULL;
	}
	return;
}

module_init(factcfg_init);
module_exit(factcfg_cleanup);

MODULE_AUTHOR("ygzhang@grandstream.com");
MODULE_DESCRIPTION("GS factory cfg driver");
MODULE_LICENSE("GPL");
